from python_reactive_ui.backends.gtk3.builtin_component import Gtk3BuiltinComponent
from python_reactive_ui import Children, Props

# fmt: off
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
# fmt: on


class ProgressBar(Gtk3BuiltinComponent):
    def _pre_init(self):
        self._set_widget(Gtk.ProgressBar.new())

    def _mount(self):
        self._mounter(self.gtk_widget)

    def _receive_props(self, new_props: Props):
        super()._receive_props(new_props)
        self.gtk_widget.set_fraction(
            new_props["fraction"] if "fraction" in new_props else 0.5
        )
        self._update_prop_with_default(
            new_props, "text", None, self.gtk_widget.set_text
        )
        self._update_prop_with_default(
            new_props, "show_text", False, self.gtk_widget.set_show_text
        )
        self._update_prop_with_default(
            new_props, "inverted", False, self.gtk_widget.set_inverted
        )

    def _receive_children(self, new_children: Children):
        pass

    def _dismount(self):
        self._dismounter(self.gtk_widget)



from python_reactive_ui.backends.gtk3.builtin_component import Gtk3BuiltinComponent
from python_reactive_ui import Children, Props

# fmt: off
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
# fmt: on


class BlackBox(Gtk3BuiltinComponent):
    def _pre_init(self, widget: Gtk.Widget):
        self._set_widget(widget)

    def _receive_props(self, new_props: Props):
        super()._receive_props(new_props)

    def _receive_children(self, _: Children):
        pass

    def _mount(self):
        self._mounter(self.gtk_widget)

    def _dismount(self):
        self._dismounter(self.gtk_widget)



from typing import Callable, Optional
from python_reactive_ui import Props
from python_reactive_ui.lib.core import BuiltinComponent, Children

# fmt: off
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
# fmt: on


class Gtk3BuiltinComponent(BuiltinComponent):
    def _set_widget(self, gtk_widget: Gtk.Widget):
        self.gtk_widget = gtk_widget

    def _post_mount(self):
        self.gtk_widget.show_all()

    def _prop_has_new_value(self, new_props: Props, prop_name: str):
        return prop_name in new_props and (
            prop_name not in self._props
            or self._props[prop_name] != new_props[prop_name]
        )

    def _prop_removed(self, new_props: Props, prop_name: str):
        return prop_name not in new_props and prop_name in self._props

    def _update_prop_with_default(
        self,
        new_props: Props,
        prop_name: str,
        default,
        updater,
        prop_mapper: Callable = lambda x: x,
    ):
        if self._prop_has_new_value(new_props, prop_name):
            updater(prop_mapper(new_props[prop_name]))
        elif self._prop_removed(new_props, prop_name):
            updater(prop_mapper(default))

    def _update_list_prop(
        self,
        new_props: Props,
        prop_name: str,
        adder,
        remover,
        prop_mapper: Callable = lambda x: x,
    ):
        new_list = new_props[prop_name] if prop_name in new_props else []
        old_list = self._props[prop_name] if prop_name in self._props else []
        for item in new_list:
            if item not in old_list:
                adder(item)
        for item in old_list:
            if item not in new_list:
                remover(item)

    def _receive_props(self, new_props: Props):
        # List props
        style_context = self.gtk_widget.get_style_context()
        self._update_list_prop(
            new_props,
            "css_classes",
            style_context.add_class,
            style_context.remove_class,
        )

        # Simple props
        self._update_prop_with_default(
            new_props, "hexpand", False, self.gtk_widget.set_hexpand
        )
        self._update_prop_with_default(
            new_props, "vexpand", False, self.gtk_widget.set_vexpand
        )
        self._update_prop_with_default(
            new_props, "sensitive", True, self.gtk_widget.set_sensitive
        )
        self._update_prop_with_default(
            new_props, "opacity", 1.0, self.gtk_widget.set_opacity
        )
        self._update_prop_with_default(
            new_props,
            "halign",
            "fill",
            self.gtk_widget.set_halign,
            self._get_align_enum,
        )
        self._update_prop_with_default(
            new_props,
            "valign",
            "fill",
            self.gtk_widget.set_valign,
            self._get_align_enum,
        )
        self._update_prop_with_default(
            new_props, "tooltip_text", None, self.gtk_widget.set_tooltip_text
        )
        self._update_prop_with_default(
            new_props,
            "size_request",
            (-1, -1),
            lambda sr: self.gtk_widget.set_size_request(*sr),
        )
        if isinstance(self.gtk_widget, Gtk.Orientable):
            self._receive_orientable_props(new_props)

    def _receive_orientable_props(self, new_props):
        self._update_prop_with_default(
            new_props,
            "orientation",
            "horizontal",
            self.gtk_widget.set_orientation,
            self._get_orientation_enum,
        )

    def _get_orientation_enum(self, orientation: str):
        match orientation:
            case "horizontal":
                return Gtk.Orientation.HORIZONTAL
            case "vertical":
                return Gtk.Orientation.VERTICAL
            case other:
                raise ValueError(f"orientation value of {other} is not supported")

    def _get_align_enum(self, value):
        match value:
            case "fill":
                return Gtk.Align.FILL
            case "start":
                return Gtk.Align.START
            case "end":
                return Gtk.Align.END
            case "center":
                return Gtk.Align.CENTER
            case "baseline":
                return Gtk.Align.BASELINE
            case other:
                raise ValueError(f"Align value {other} not recognized.")



from python_reactive_ui import Component
from python_reactive_ui.lib.core import Root

# fmt: off
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
# fmt: on


class Gtk3Root(Root):
    def __init__(self, root: Gtk.Container):
        self.root = root
        self.tree = None

    def render(self, component: Component):
        component.mount(self.root.add, self.root.remove)
        self.root.show_all()


def create_root(root):
    return Gtk3Root(root)



import logging
import logging.handlers
from python_reactive_ui.backends.gtk3 import create_root
from python_reactive_ui.backends.gtk3.builtin import Box
from python_reactive_ui.backends.gtk3.builtin import Label
from python_reactive_ui.backends.gtk3.builtin import Button
from python_reactive_ui.backends.gtk3.builtin import ProgressBar
from python_reactive_ui.backends.gtk3.builtin.blackbox import BlackBox
from python_reactive_ui import (
    Children,
    Props,
    use_state,
    Component,
)

# fmt: off
import gi

from python_reactive_ui.lib.hooks import use_effect

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
# fmt: on

testlabel = Gtk.Label.new("blackbox")


class Incremental(Component):
    def _render(self, props: Props, children: Children):
        counter, set_counter = use_state(self, 0)
        use_effect(self, lambda: print("Ayoooo"), [counter == 3])

        box_children = [
            Button(
                {
                    "on_click": lambda _: set_counter(counter + 1),
                    "vexpand": counter % 2 == True,
                },
                [Label({"text": "+1"})],
            ),
            Label(
                {"text": f"Counted {counter} clicks", "css_classes": ["test-class"]},
            ),
            ProgressBar({"fraction": counter / 50}),
            # BlackBox({}, widget=testlabel),
        ]

        if counter > 5 and counter < 10:
            box_children.append(Label({"text": f"Loads of clicks!"}))

        elem = Box({"orientation": "vertical"}, box_children)
        return elem


def test():
    setup_logging()
    root = create_root(Gtk.Window.new(Gtk.WindowType.TOPLEVEL))
    elem = Incremental(dict())
    # elem._print_on_render = True
    root.render(elem)
    Gtk.main()


def setup_logging():
    logging.basicConfig(
        level=logging.INFO,
        style="{",
        format="[{levelname}({name}):{filename}:{funcName}] {message}",
    )
    root_logger = logging.getLogger()
    sys_handler = logging.handlers.SysLogHandler(address="/dev/log")
    root_logger.addHandler(sys_handler)



from python_reactive_ui.backends.gtk3.builtin.box import Box
from python_reactive_ui.backends.gtk3.builtin.button import Button
from python_reactive_ui.backends.gtk3.builtin.label import Label
from python_reactive_ui.backends.gtk3.builtin.progress_bar import ProgressBar



from python_reactive_ui.backends.gtk3.builtin_component import Gtk3BuiltinComponent
from python_reactive_ui import Children, Props

# fmt: off
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
# fmt: on


class Box(Gtk3BuiltinComponent):
    def _pre_init(self):
        self._set_widget(Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0))

    def _receive_props(self, new_props: Props):
        super()._receive_props(new_props)
        self.gtk_widget.set_spacing(
            new_props["spacing"] if "spacing" in new_props else 0
        )

    def _receive_children(self, new_children: Children):
        actions = self._compare_children(self._children, new_children)
        self._perform_actions(actions, self.gtk_widget.add, self.gtk_widget.remove)
        self.gtk_widget.show_all()

    def _mount(self):
        self._mounter(self.gtk_widget)

    def _dismount(self):
        self._dismounter(self.gtk_widget)



from enum import Enum, auto
from collections import deque
from typing import Callable, Deque, Dict, List, Optional, Tuple
from itertools import zip_longest
import inspect


class Root:
    def render(self, component: "Component"):
        raise NotImplementedError()


Props = Dict
Children = List["Component"]


class ActionType(Enum):
    SWAP = auto()
    MOUNT = auto()
    DISMOUNT = auto()
    UPDATE = auto()
    ADD_CHILD = auto()
    REMOVE_CHILD = auto()


Action = Tuple[ActionType, Tuple]
Actions = List[Action]


class Component:
    def __init__(self, props: Props, children: Optional[Children] = None, **kwargs):
        self._pre_init(**kwargs)
        self._init_empty()
        self._init_params(props, children)
        self._post_init(**kwargs)

    def _pre_init(self, **kwargs):
        pass

    def _post_init(self, **kwargs):
        pass

    def _init_empty(self):
        self._children: Children = []
        self._props: Props = dict()
        self._render_result: Optional["Component"] = None
        self._render_parent: Optional["Component"] = None

        # State
        self._state: Dict = dict()
        self._state_counter = 0

        # Effects
        self._effects: Dict = dict()
        self._effect_counter = 0

        # Other
        self._print_on_render = False
        self._is_mounted = False
        self._queued_operations: Deque = deque()

    def _init_params(
        self,
        props: Props,
        children: Optional[Children] = None,
    ):
        _children: Children = [] if children is None else children
        self.receive_props(props)
        self.receive_children(_children)
        self._render_close()

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return f"<{type(self).__name__}>"

    def _enqueue_operation(self, operation: Callable):
        self._queued_operations.append(operation)

    def _run_operations(self):
        while self._queued_operations:
            op = self._queued_operations.popleft()
            op()

    def render(self) -> "Component":
        if self._print_on_render:
            print_tree(self)
        self._state_counter = 0
        self._effect_counter = 0
        result = self._render(self._props, self._children)
        result._render_parent = self
        return result

    def _render(self, props: Props, children: Children) -> "Component":
        raise NotImplementedError()

    def _render_close(self):
        if not isinstance(self, BuiltinComponent):
            self._render_result = self.render()
            self._render_result._render_close()

    def trigger_rerender(
        self, mounter=Optional[Callable], dismounter=Optional[Callable]
    ):
        mounter = mounter if (mounter is not None) else self._mounter
        dismounter = dismounter if (dismounter is not None) else self._dismounter

        if not isinstance(self, BuiltinComponent):
            new_element = self.render()
            actions = self._compare(self._render_result, new_element)
            self._perform_actions(actions, mounter, dismounter)  # type: ignore

        self._run_operations()

    def _compare_children(
        self,
        old_children: List["Component"],
        new_children: List["Component"],
    ) -> Actions:
        actions = []
        for i, (old, new) in enumerate(zip_longest(old_children, new_children)):
            curr_actions = self._compare(old, new)
            if len(curr_actions) == 1:
                [action] = curr_actions
                match action:
                    case (ActionType.MOUNT, (new,)):
                        actions.extend([action, (ActionType.ADD_CHILD, (new,))])
                    case (ActionType.DISMOUNT, (old,)):
                        actions.extend([action, (ActionType.REMOVE_CHILD, (old,))])
                    case (ActionType.SWAP, (old, new)):
                        actions.extend([action, (ActionType.ADD_CHILD, (old, i))])
                    case other:
                        actions.append(other)
        return actions

    def _compare(
        self,
        old: Optional["Component"],
        new: Optional["Component"],
    ) -> Actions:
        if new is not None:
            if old is None:
                return [(ActionType.MOUNT, (new,))]

            elif not type(old) == type(new):
                return [(ActionType.SWAP, (old, new))]
            else:
                if old._props != new._props:
                    return [(ActionType.UPDATE, (old, new))]
                if len(self._compare_children(old._children, new._children)) > 0:
                    return [(ActionType.UPDATE, (old, new))]
        else:
            if old is not None:
                return [(ActionType.DISMOUNT, (old,))]
        return []

    def _perform_actions(
        self, actions: Actions, mounter: Callable, dismounter: Callable
    ):
        for action in actions:
            match action:
                case (ActionType.MOUNT, (new,)):
                    if new._is_mounted:
                        new.dismount()
                    new.mount(mounter, dismounter)
                case (ActionType.DISMOUNT, (old,)):
                    old.dismount()
                case (ActionType.SWAP, (old, new)):
                    if old._render_parent is not None:
                        assert old._render_parent._render_child is old
                        assert new._render_parent is not None
                        assert new._render_parent is old._render_parent
                        old._render_parent._render_child = new
                    old.dismount()
                    new.mount(mounter, dismounter)
                case (ActionType.UPDATE, (old, new)):
                    old.receive_props(new._props)
                    old.receive_children(new._children)
                    old.trigger_rerender()
                case (ActionType.ADD_CHILD, (new,)):
                    self._add_child(new)
                case (ActionType.ADD_CHILD, (new, i)):
                    self._add_child(new, i)
                case (ActionType.REMOVE_CHILD, (old,)):
                    self._remove_child(old)

    def receive_props(self, new_props: Props):
        self._pre_receive_props(new_props)
        self._receive_props(new_props)
        self._post_receive_props(new_props)

    def _pre_receive_props(self, new_props: Props):
        pass

    def _receive_props(self, new_props: Props):
        pass

    def _post_receive_props(self, new_props: Props):
        self._props = new_props

    def receive_children(self, new_children: Children):
        self._receive_children(new_children)

    def _receive_children(self, new_children: Children):
        self._children = new_children

    def mount(
        self,
        mounter: Optional[Callable] = None,
        dismounter: Optional[Callable] = None,
    ):
        if mounter is not None:
            self._mounter = mounter
        if dismounter is not None:
            self._dismounter = dismounter
        assert mounter is not None
        assert dismounter is not None
        assert not self._is_mounted
        self._is_mounted = True
        self._pre_mount()
        self._mount()
        self._post_mount()
        self._run_operations()

    def _pre_mount(self):
        pass

    def _mount(self):
        assert self._render_result is not None
        return self._render_result.mount(self._mounter, self._dismounter)

    def _post_mount(self):
        pass

    def dismount(self):
        assert self._is_mounted
        assert self._dismounter is not None
        self._is_mounted = False
        self._dismount()

    def _dismount(self):
        assert self._render_result is not None
        return self._render_result.dismount()

    def _add_child(self, child: "Component", index: Optional[int] = None):
        if index is not None and len(self._children) > index:
            self._children[index] = child
        else:
            self._children.append(child)

    def _remove_child(self, child: "Component"):
        self._children.remove(child)


class BuiltinComponent(Component):
    def _mount(self):
        raise NotImplementedError()

    def _receive_props(self, new_props: Props):
        raise NotImplementedError()

    def _receive_children(self, new_children: Children):
        raise NotImplementedError()

    def _dismount(self):
        raise NotImplementedError()


def print_actions(actions):
    ss = "[\n"
    for action in actions:
        (action_type, stuff) = action
        s = f"({str(action_type)}"
        for a in stuff:
            s += f", {a}"
        s += ")"
        ss += f"\t{s}\n"
    ss += "]"
    return ss


def print_tree(component: Component, depth=0):
    print("| " * depth + f"{component}")
    if not isinstance(component, BuiltinComponent):
        print("| " * depth + "|" + f"Renders:")
        if component._render_result:
            print_tree(component._render_result, depth=depth + 1)
    if len(component._children) > 0:
        print("| " * depth + "|" + f"Children:")
        for child in component._children:
            print_tree(child, depth=depth + 1)



from typing import Callable, Optional
from python_reactive_ui.backends.gtk3.builtin_component import Gtk3BuiltinComponent
from python_reactive_ui import Children, Props

# fmt: off
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
# fmt: on


class Button(Gtk3BuiltinComponent):
    def _pre_init(self):
        self._set_widget(Gtk.Button.new())
        self.signal_handlers = dict()

    def _receive_props(self, new_props: Props):
        super()._receive_props(new_props)
        self._update_on_click(
            new_props["on_click"] if "on_click" in new_props else None
        )

    def _receive_children(self, new_children: Children):
        if len(new_children) > 1:
            raise ValueError("button only supports a single child")
        actions = self._compare_children(self._children, new_children)
        self._perform_actions(actions, self.gtk_widget.add, self.gtk_widget.remove)

    def _update_on_click(self, on_click: Optional[Callable]):
        if "clicked" in self.signal_handlers:
            self.gtk_widget.disconnect(self.signal_handlers["clicked"])
            self.signal_handlers.pop("clicked")
        if on_click:
            self.signal_handlers["clicked"] = self.gtk_widget.connect(
                "clicked", on_click
            )

    def _dismount(self):
        self._dismounter(self.gtk_widget)

    def _mount(self):
        self._mounter(self.gtk_widget)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['python_reactive_ui',
 'python_reactive_ui.backends',
 'python_reactive_ui.backends.gtk3',
 'python_reactive_ui.backends.gtk3.builtin',
 'python_reactive_ui.lib']

package_data = \
{'': ['*']}

install_requires = \
['PyGObject>=3.42.2,<4.0.0']

entry_points = \
{'console_scripts': ['example = python_reactive_ui.main:test']}

setup_kwargs = {
    'name': 'python-reactive-ui',
    'version': '0.2.10',
    'description': '',
    'long_description': '# python-reactive-ui\n\nHave you ever wanted to use React, only to stumble upon the realization that\nyour project uses Python and GTK for the UI? Grieve no more, my friend, for this\nproject seeks to save you from your choices.\n\nAttach components to arbitrary GTK containers.\n',
    'author': 'Manuel Brea',
    'author_email': 'm.brea.carreras@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from collections.abc import Callable
from typing import List
from python_reactive_ui import Component


def use_state(component: Component, initial_value=None):
    counter = component._state_counter

    def set_func(new_value):
        component._state[counter] = new_value
        component.trigger_rerender()

    if component._state_counter not in component._state:
        component._state[component._state_counter] = initial_value
    elem = component._state[component._state_counter]
    component._state_counter += 1
    return elem, set_func


def use_effect(component: Component, callback: Callable, deps: List):
    if (
        component._effect_counter not in component._effects
        or component._effects[component._effect_counter][0] != deps
    ):
        component._effects[component._effect_counter] = (deps, callback)
        component._enqueue_operation(callback)
    component._effect_counter += 1
